!
! Purpose: tests the correctness and performance of blocked in-place
!  transposition algorithms.
!
! Usage: ./test m n me ne L verify numrep
!
!  m, n:    size of matrix, m x n
!
!  me, ne:  size of transposition problem, me x ne
!
!  L:       block size, L contiguous elements (note that me*ne*L must divide m*n)
!
!  verify:  verify result (yes/no)
!
!  numrep:  number of repetitions, select minimum execution time
!
program test
   use ipsfc
   use omp_lib

   implicit none

   integer :: m, n, p = 1
   double precision, dimension(:,:), allocatable :: A
   double precision :: t1, t2, Tp, Ttest, Tver
   character(100) :: arg
   integer :: narg

   integer :: i, j, ii, jj, k, kk, ok, me, ne, L, nprob
   integer :: d4, d3, d2, d1, x4, x3, x2, nrep = 1, rep 
   logical :: verifyresult = .false.

!$omp parallel 
!$omp master
   p = omp_get_num_threads()
!$omp end master
!$omp end parallel

   ! parse command line arguments
   narg = command_argument_count()
   if( narg == 7 )then
      call get_command_argument(1, arg)
      read(arg,'(i8)') m
      call get_command_argument(2, arg)
      read(arg,'(i8)') n
      call get_command_argument(3, arg)
      read(arg,'(i8)') me
      call get_command_argument(4, arg)
      read(arg,'(i8)') ne
      call get_command_argument(5, arg)
      read(arg,'(i8)') L
      call get_command_argument(6, arg)
      if( arg == 'yes' ) verifyresult = .true.
      call get_command_argument(7, arg)
      read(arg,'(i8)') nrep
   else
      call error('Usage: m n me ne L verify numrep')
   end if

   ! check inputs
   if( mod(m * n, me * ne * L) /= 0 )then
      call error('test: problem size does not match matrix size')
   end if
   nprob = (m * n) / (me * ne * L)

   ! allocate storage
   allocate(A(m,n))

   do rep = 1, nrep
      
      Ttest = gettime()

      ! fill the matrix with specific numbers
      call generate_CM(m, n, A(1,1), m)
      !!!write(*,*) 'Initial A:'
      !!!call print_CM(m, n, A(1,1), m)
      
      ! --- START CLOCK ---
      t1 = gettime()

      ! do the transposition
      call inplace_transposition_driver(p, m, n, A(1,1), nprob, me, ne, L)

      ! --- STOP CLOCK ---
      t2 = gettime()
      Tp = t2 - t1
      
      !!!write(*,*) 'Resulting A:'
      !!!call print_CM(n, m, A(1,1), n)
      
      ! verify the result
      Tver = gettime()
      if( verifyresult )then
         d4 = nprob
         d3 = ne
         d2 = me
         d1 = L
         ok = 1
         do k = 0, m*n - 1
            i = mod(k, m)
            j = k / m
            x4 = k / (d3 * d2 * d1)
            x3 = mod(k, d3 * d2 * d1) / (d2 * d1)
            x2 = mod(k, d2 * d1) / (d1)
            kk = x4*d3*d2*d1 + x2*d3*d1 + x3*d1 + mod(k, d1)
            ii = 1 + mod(kk, m)
            jj = 1 + kk / m
            if( A(ii, jj) /= k + 1 )then
               ok = 0
               exit
            end if
         end do
      else
         ok = 2
      end if
      Tver = gettime() - Tver
      
      Ttest = gettime() - Ttest

      ! print results
      write(*,'(a,11a10)') '# ', 'nth', 'd4', 'd2(me)', 'd3(ne)', &
         'd1(L)', 'Tp', 'Tp/el', '(mem: MB)', '(Ttest)', '(Tver)', &
         'verified?'
      write(*,'(a2,5i10,f10.6,e10.3,i10,2f10.6,1i10)',advance='no') '', &
         p, nprob, me, ne, L, &
         Tp, Tp / (nprob * me * ne * L), &
         (nprob * me * ne * L) / 1024 / 128, Ttest, Tver, ok
      write(*,*)
   end do ! end do rep 

   deallocate(A)

end program test
